//---------------------------------------------------------------------------
// entropykey.cpp
//
// Windows DLL for Simtec Entropy Key - www.entropykey.co.uk
//
// Version 1, 21 Oct 2012, Author: Nicolas Lynch-Aird
// Version 2, 22 Oct 2012, Author: Nicolas Lynch-Aird
//			  Ported to MS Visual Studio 2010 Express
// Version 3, 30 Oct 2012, Authors: Nicolas Lynch-Aird, Christopher Mioni
//			  Improvements to solution/project structure and README file
// Version 4, 01 Nov 2012, Author: Nicolas Lynch-Aird
//			  Added use of text file to hold list of entropy key details
//
// NOTE: Set character set to "Use Multi-Byte Character Set" in Configuration Properties
// 
// This software is released for free use, modification, and distribution. It
// has not been formally tested and is provided with no warranties whatsoever.
//
//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setupapi.h>
#pragma comment (lib, "setupapi.lib")
#include <initguid.h>
#include <time.h>

// These files can be found in the device directory of Simtec's software set ekeyd-1.1.4.tar.gz:
//   - frames/frame.c
//   - frames/frame.h
//   - frames/pem.c
//   - frames/pem.h
//   - skein/brg_endian.h
//   - skein/brg_types.h
//   - skein/skein.c
//   - skein/skein.h
//   - skein/skein_block.c
//   - skein/skein_iv.h
//   - skein/skein_port.h
//   - skeinwrap.c
//   - skeinwrap.h
//   
#include "frames/pem.c"
#include "skein/skein_block.c"
#include "skein/skein.c"
#include "skeinwrap.c"

#define	BUILDING_DLL
#include "entropykey.h"

#define EKEY_LIST_FILE		"ekey_list.txt"
#define	LINELENGTH			100
#define VID_PID				"vid_20df&pid_0001"
#define PROTOCOL_VERSION	"AA"
#define	COM_OUT_LEN			20
#define PACKET_LEN			64
#define	READ_ATTEMPTS		50
#define NONCE_LEN			12

#pragma warning(disable:4996)

// This is the GUID for the USB device class (A5DCBF10-6530-11D2-901F-00C04FB951ED)
DEFINE_GUID(GUID_DEVINTERFACE_USB_DEVICE, 0xA5DCBF10L, 0x6530, 0x11D2, 0x90, 0x1F, 0x00, 0xC0, 0x4F, 0xB9, 0x51, 0xED);

struct ekeylist
{
	char	serial_number[20];	// text version
	char	master_key[50];     // text version
	char	serial_frame[70];
};

ekeylist 	ekey_list[MAX_KEYS];

int	ekey_list_count = 0;

struct ekeyfound
{
	int     	ekey_list_num;
	BYTE		serial_num[12];		// binary version
	BYTE		master_key[33];     // binary version - need an extra byte for pem64_decode from text string
	BYTE		long_term_key[32];  // binary version
	BYTE		session_key[32];    // binary version
	int			sequence_num;
	BYTE		nonce[NONCE_LEN];
	char		com_port[10];
	HANDLE		hCom;
	OVERLAPPED	com_overlapped;
	char		input_bytes[PACKET_LEN];   // parameters for reading data from device
	DWORD		input_count;
};

ekeyfound		ekey[MAX_KEYS];

int				keys_found = 0;

HANDLE			com_event[MAX_KEYS];

HANDLE 			ekey_thread;
BOOL			stop_thread;

BYTE			pool[POOLSIZE];
int				pool_write_index = 0;		// index for next write
int				pool_read_index = 0;        // index for next read
int				pool_bytes_available = 0;
BOOL			pool_write_flag = FALSE;
BOOL			pool_read_flag = FALSE;

//---------------------------------------------------------------------------

void GetTimestamp(char* datetimestr)
{
  SYSTEMTIME	date_time;

  GetSystemTime(&date_time);
  sprintf(datetimestr, "%02d/%02d/%04d %02d:%02d:%02d - ", date_time.wDay, date_time.wMonth,
			date_time.wYear, date_time.wHour, date_time.wMinute, date_time.wSecond);
}

void LogFrame(char* com_port, char* input_bytes, DWORD input_count)
{
	DWORD 	i;
	char    timestamp[25];

	if(ekey_log_fptr != NULL)
	{
		GetTimestamp(timestamp);
		fprintf(ekey_log_fptr, "%s%s: ", timestamp, com_port);
		for(i=0;i<input_count;i++) fprintf(ekey_log_fptr, "%c", input_bytes[i]);
		// don't need to include new-line as included at end of frame
	}
}

void LogError(char* com_port, char* errmsg)
{
	char    timestamp[25];

	if(ekey_log_fptr != NULL)
	{
		GetTimestamp(timestamp);
		fprintf(ekey_log_fptr, "%s%s: %s\n", timestamp, com_port, errmsg);
	}
}

void LogData(BYTE* entropy, int e_count)
{
	int		i;

	if((ekey_dat_fptr != NULL) && (e_count > 0))
	{
		for(i=0;i<e_count-1;i++) fprintf(ekey_dat_fptr, "%02x ", entropy[i]);
		fprintf(ekey_dat_fptr, "%02x\n", entropy[e_count-1]);
    }
}

void LoadKeyDetails()
{
	FILE*		keyfile_fptr;
	char		keyfile_str[LINELENGTH];
	BYTE		snum[12], skey[32], macbuf[32];
	EKeySkein   skein;
	char		timestamp[25];
	int			i;

	ekey_list_count = 0;
	// initialise session key to zero
	for(i=0;i<32;i++) skey[i] = 0;
	// open entropy key list file
	keyfile_fptr = fopen(EKEY_LIST_FILE, "r");
	if(keyfile_fptr != NULL)
	{
		while(fgets(keyfile_str, LINELENGTH, keyfile_fptr) != NULL)
			if(strncmp(keyfile_str, "#", 1) != 0) // ignore lines starting with #
				// read in the next set of entropy key details
				if(sscanf(keyfile_str, "%s %s", ekey_list[ekey_list_count].serial_number, ekey_list[ekey_list_count].master_key) == 2)
				{
					// calculate the serial number frame
					// first part of frame
					strcpy(ekey_list[ekey_list_count].serial_frame, "* S!");
					strcat(ekey_list[ekey_list_count].serial_frame, PROTOCOL_VERSION);
					strcat(ekey_list[ekey_list_count].serial_frame, ekey_list[ekey_list_count].serial_number);
					strcat(ekey_list[ekey_list_count].serial_frame, "                                "); // 32 spaces
					// convert serial number to binary
					pem64_decode_bytes(ekey_list[ekey_list_count].serial_number, 16, snum);
					// calculate MAC
					PrepareSkein(&skein, snum, skey, EKEY_SKEIN_PERSONALISATION_PMS);
					Skein_256_Update(&skein, (BYTE*) ekey_list[ekey_list_count].serial_frame + 2, 52);
					Skein_256_Final_Pad(&skein, macbuf);
					pem64_encode_bytes(macbuf, 3, ekey_list[ekey_list_count].serial_frame + 54);
					pem64_encode_bytes(macbuf + 29, 3, ekey_list[ekey_list_count].serial_frame + 58);
					// finish the frame
					ekey_list[ekey_list_count].serial_frame[62] = '\r';
					ekey_list[ekey_list_count].serial_frame[63] = '\n';
					ekey_list[ekey_list_count].serial_frame[64] = 0; // terminating null (in case of display)
					// increment count
					ekey_list_count++;
				}
				else
					if(ekey_log_fptr != NULL)
					{
						GetTimestamp(timestamp);
						fprintf(ekey_log_fptr, "%s%s: Format Error > %s\n", timestamp, EKEY_LIST_FILE, keyfile_str);
					}

		fclose(keyfile_fptr);
	}
	else
		if(ekey_log_fptr != NULL)
		{
			GetTimestamp(timestamp);
			fprintf(ekey_log_fptr, "%sUnable to open %s\n", timestamp, EKEY_LIST_FILE);
		}
}

void NewNonce(int devnum)
{
	int		i;
	for(i=0;i<NONCE_LEN;i++)
		ekey[devnum].nonce[i] = rand() % 256;
}

BOOL ValidMAC(int devnum, char* input_bytes, DWORD input_count)
{
	BOOL	return_value = FALSE;		// set for fail by default

	EKeySkein   skein;
	BYTE	macbuf[32];
	char	MAC[10];

	// first check packet length and framing
	if((input_count == PACKET_LEN) && (strncmp(input_bytes, "* ", 2) == 0)
		&& (input_bytes[PACKET_LEN-2] == 13) && (input_bytes[PACKET_LEN-1] == 10))
	{
		// calculate MAC
		PrepareSkein(&skein, ekey[devnum].serial_num, ekey[devnum].session_key, EKEY_SKEIN_PERSONALISATION_PMS);
		Skein_256_Update(&skein, (BYTE*) input_bytes + 2, 52);
		Skein_256_Final_Pad(&skein, macbuf);
		pem64_encode_bytes(macbuf, 3, MAC);
		pem64_encode_bytes(macbuf + 29, 3, MAC + 4);
		// compare with received MAC
		return_value = (strncmp(input_bytes + 54, MAC, 8) == 0);
	}

	return return_value;
}

BOOL LongTermKey(int devnum)
{
	BOOL   	return_value = FALSE;		// set for fail by default

	EKeySkein   skein;
	BYTE	macbuf[32];
	int		pem64_nonce_len;
	int		sub_code;
	BYTE	rx_data[40];
	int		rx_count;

	char	output_bytes[COM_OUT_LEN];  // parameters for writing data to device
	DWORD	output_count;
	DWORD	request_count;
	char	input_bytes[PACKET_LEN];   // parameters for reading data from device
	DWORD	input_count;
	int		read_attempt;

	// create nonce
	NewNonce(devnum);
	// create nonce MAC
	PrepareSkein(&skein, ekey[devnum].serial_num, ekey[devnum].master_key, EKEY_SKEIN_PERSONALISATION_LKMS);
	Skein_256_Update(&skein, ekey[devnum].nonce, NONCE_LEN);
	Skein_256_Final_Pad(&skein, macbuf);
	// construct output string
	output_bytes[0] = 'M';
	pem64_encode_bytes(macbuf, 3, output_bytes + 1);
	pem64_encode_bytes(macbuf + 29, 3, output_bytes + 5);
	request_count = 9;
	// send nonce MAC
	WriteFile(ekey[devnum].hCom, output_bytes, request_count, &output_count, &ekey[devnum].com_overlapped);
	GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &output_count, TRUE);
	// retrieve long-term-rekey MAC acknowledge frame
	read_attempt = 0;
	do
	{
		// read next packet
		ReadFile(ekey[devnum].hCom, input_bytes, PACKET_LEN, &input_count, &ekey[devnum].com_overlapped);
		GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &input_count, TRUE);
		LogFrame(ekey[devnum].com_port, input_bytes, input_count);
	}
	while((++read_attempt < READ_ATTEMPTS) && (strncmp(input_bytes, "* M>", 4) != 0));
	if(read_attempt < READ_ATTEMPTS)
	{
		if(ValidMAC(devnum, input_bytes, input_count))
		{
			// construct output string
			output_bytes[0] = 'L';
			pem64_encode_bytes(ekey[devnum].nonce, NONCE_LEN, output_bytes + 1);
			pem64_nonce_len = pem64_encode_bytes_length(NONCE_LEN);
			output_bytes[pem64_nonce_len + 1] = '.';
			request_count = pem64_nonce_len + 2;
			// send nonce
			WriteFile(ekey[devnum].hCom, output_bytes, request_count, &output_count, &ekey[devnum].com_overlapped);
			GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &output_count, TRUE);
			// retrieve long-term-rekey acknowledge frame
			read_attempt = 0;
			do
			{
				// read next packet
				ReadFile(ekey[devnum].hCom, input_bytes, PACKET_LEN, &input_count, &ekey[devnum].com_overlapped);
				GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &input_count, TRUE);
				LogFrame(ekey[devnum].com_port, input_bytes, input_count);
			}
			while((++read_attempt < READ_ATTEMPTS) && (strncmp(input_bytes, "* L!", 4) != 0));
			if(read_attempt < READ_ATTEMPTS)
			{
				if(ValidMAC(devnum, input_bytes, input_count))
				{
					// check sub-code matches length of nonce
					sub_code = pem64_decode_12bits(input_bytes + 4);
					if(sub_code == NONCE_LEN)
					{
						// decode to get binary data load
						rx_count = pem64_decode_bytes(input_bytes + 6, 48, rx_data);
						// calculate new long-term key
						PrepareSkein(&skein, ekey[devnum].serial_num, ekey[devnum].master_key, EKEY_SKEIN_PERSONALISATION_LRS);
						Skein_256_Update(&skein, rx_data, rx_count);
						Skein_256_Update(&skein, ekey[devnum].nonce, NONCE_LEN);
						Skein_256_Final(&skein, ekey[devnum].long_term_key);
						// success
						return_value = TRUE;
					} // if(sub_code == NONCE_LEN)
                    else LogError(ekey[devnum].com_port, "Sub-code <> nonce length");
				} // if(ValidMAC(devnum, input_bytes, input_count))
				else LogError(ekey[devnum].com_port, "Invalid frame structure or MAC");
			} // if(read_attempt < READ_ATTEMPTS)
			else LogError(ekey[devnum].com_port, "Failed to retrieve Long-Term-Rekey Acknowledge Frame");
		} // if(ValidMAC(devnum, input_bytes, input_count))
		else LogError(ekey[devnum].com_port, "Invalid frame structure or MAC");
	} // if(read_attempt < READ_ATTEMPTS)
	else LogError(ekey[devnum].com_port, "Failed to retrieve Long-Term-Rekey MAC Acknowledge Frame");

	return return_value;
}

BOOL SessionKey(int devnum)
{
	BOOL   	return_value = FALSE;		// set for fail by default

	EKeySkein   skein;
	int		pem64_nonce_len;
	int		sub_code;
	BYTE	rx_data[40];
	int		rx_count;

	char	output_bytes[COM_OUT_LEN];  // parameters for writing data to device
	DWORD	output_count;
	DWORD	request_count;
	char	input_bytes[PACKET_LEN];   // parameters for reading data from device
	DWORD	input_count;
	int		read_attempt;

	// create nonce
	NewNonce(devnum);
	// construct output string
	output_bytes[0] = 'K';
	pem64_encode_bytes(ekey[devnum].nonce, NONCE_LEN, output_bytes + 1);
	pem64_nonce_len = pem64_encode_bytes_length(NONCE_LEN);
	output_bytes[pem64_nonce_len + 1] = '.';
	request_count = pem64_nonce_len + 2;
	// send nonce
	WriteFile(ekey[devnum].hCom, output_bytes, request_count, &output_count, &ekey[devnum].com_overlapped);
	GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &output_count, TRUE);
	// retrieve session re-key frame
	read_attempt = 0;
	do
	{
		// read next packet
		ReadFile(ekey[devnum].hCom, input_bytes, PACKET_LEN, &input_count, &ekey[devnum].com_overlapped);
		GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &input_count, TRUE);
		LogFrame(ekey[devnum].com_port, input_bytes, input_count);
	}
	while((++read_attempt < READ_ATTEMPTS) && (strncmp(input_bytes, "* K!", 4) != 0));
	if(read_attempt < READ_ATTEMPTS)
	{
		if(ValidMAC(devnum, input_bytes, input_count))
		{
			// check sub-code matches length of nonce
			sub_code = pem64_decode_12bits(input_bytes + 4);
			if(sub_code == NONCE_LEN)
			{
				// decode to get binary data load
				rx_count = pem64_decode_bytes(input_bytes + 6, 48, rx_data);
				// calculate new session key
				PrepareSkein(&skein, ekey[devnum].serial_num, ekey[devnum].long_term_key, EKEY_SKEIN_PERSONALISATION_RS);
				Skein_256_Update(&skein, rx_data, rx_count);
				Skein_256_Update(&skein, ekey[devnum].nonce, NONCE_LEN);
				Skein_256_Final(&skein, ekey[devnum].session_key);
				// reset expected sequence number
				ekey[devnum].sequence_num = 0;
				// success
				return_value = TRUE;
			} // if(sub_code == NONCE_LEN)
			else LogError(ekey[devnum].com_port, "Sub-code <> nonce length");
		} // if(ValidMAC(devnum, input_bytes, input_count))
		else LogError(ekey[devnum].com_port, "Invalid frame structure or MAC");
	} // if(read_attempt < READ_ATTEMPTS)
	else LogError(ekey[devnum].com_port, "Failed to retrieve Session Re-key Frame");

	return return_value;
}

int FindKeys()
{
	// The identification of the COM port associated with a connected device
	// is based on code provided by Velleman support

	HDEVINFO                         hDevInfo;
	SP_DEVICE_INTERFACE_DATA         DevIntfData;
	PSP_DEVICE_INTERFACE_DETAIL_DATA DevIntfDetailData;
	SP_DEVINFO_DATA                  DevData;

	DWORD	dwSize, dwType, dwMemberIdx;
	HKEY	hKey;
	BYTE	lpData[10];
	char	comstr[15];

	int		devnum = 0;
	int     listnum;
	BOOL	found;

	COMMTIMEOUTS	com_port_timeouts;

	char	input_bytes[PACKET_LEN];   // parameters for reading data from device
	DWORD	input_count;
	int		read_attempt;
	int		i;

	// We will try to get device information set for all USB devices that have a
	// device interface and are currently present on the system (plugged in).
	hDevInfo = SetupDiGetClassDevs(&GUID_DEVINTERFACE_USB_DEVICE, NULL, 0, DIGCF_DEVICEINTERFACE | DIGCF_PRESENT);
	if(hDevInfo != INVALID_HANDLE_VALUE)
	{
		// Prepare to enumerate all device interfaces for the device information
		// set that we retrieved with SetupDiGetClassDevs(..)
		DevIntfData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
		dwMemberIdx = 0;
		// Next, we will keep calling this SetupDiEnumDeviceInterfaces(..) until this
		// function causes GetLastError() to return  ERROR_NO_MORE_ITEMS. With each
		// call the dwMemberIdx value needs to be incremented to retrieve the next
		// device interface information.
		SetupDiEnumDeviceInterfaces(hDevInfo, NULL, &GUID_DEVINTERFACE_USB_DEVICE, dwMemberIdx, &DevIntfData);
		while(GetLastError() != ERROR_NO_MORE_ITEMS)
		{
			// As a last step we will need to get some more details for each
			// set of device interface information we are able to retrieve. This
			// device interface detail gives us the information we need to identify
			// the device (VID/PID), and decide if it's useful to us. It will also
			// provide a DEVINFO_DATA structure which we can use to know the serial
			// port name for a virtual com port.
			DevData.cbSize = sizeof(SP_DEVINFO_DATA);
			// Get the required buffer size. Call SetupDiGetDeviceInterfaceDetail with
			// a NULL DevIntfDetailData pointer, a DevIntfDetailDataSize
			// of zero, and a valid RequiredSize variable. In response to such a call,
			// this function returns the required buffer size at dwSize.
			SetupDiGetDeviceInterfaceDetail(hDevInfo, &DevIntfData, NULL, 0, &dwSize, NULL);
			// Allocate memory for the DeviceInterfaceDetail struct. Don't forget to
			// deallocate it later!
			DevIntfDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwSize);
			DevIntfDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			if(SetupDiGetDeviceInterfaceDetail(hDevInfo, &DevIntfData, DevIntfDetailData, dwSize, &dwSize, &DevData))
			{
				// Finally we can start checking if we've found a useable device,
				// by inspecting the DevIntfDetailData->DevicePath variable.
				// The DevicePath looks something like this:
				//
				// \\?\usb#vid_04d8&pid_0033#5&19f2438f&0&2#{a5dcbf10-6530-11d2-901f-00c04fb951ed}
				//
				// As you can see it contains the VID/PID for the device, so we can check
				// for the right VID/PID with string handling routines.
				//
				// The VID for the entropy key is 20df and the PID is 0001
				//
				if(strstr(DevIntfDetailData->DevicePath, VID_PID) != NULL)
				{
					// To find out the serial port for our device, we'll need to check the registry:
					hKey = SetupDiOpenDevRegKey(hDevInfo, &DevData, DICS_FLAG_GLOBAL, 0, DIREG_DEV, KEY_READ);
					dwType = REG_SZ;
					dwSize = sizeof(lpData);
					RegQueryValueEx(hKey, "PortName", NULL, &dwType, lpData, &dwSize);
					RegCloseKey(hKey);
					// Eureka!
					// Next step is to send a reset command to the key and retrieve its serial number
					strcpy_s(ekey[devnum].com_port, sizeof(ekey[devnum].com_port), (char*) lpData);
					// create a 'file' for read/write access to the key through its serial COM port
					strcpy(comstr, "\\\\.\\");
					strcat(comstr, ekey[devnum].com_port);
					ekey[devnum].hCom = CreateFile(comstr, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
					if(ekey[devnum].hCom != INVALID_HANDLE_VALUE)
					{
						// Create Manual Reset Event and OVERLAPPED objects for read/write operations
						ekey[devnum].com_overlapped.Offset = 0;
						ekey[devnum].com_overlapped.OffsetHigh = 0;
						ekey[devnum].com_overlapped.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
						com_event[devnum] = ekey[devnum].com_overlapped.hEvent;
						// Set port timeouts for read/write operations (disable timeouts)
						com_port_timeouts.ReadIntervalTimeout = 0;
						com_port_timeouts.ReadTotalTimeoutMultiplier = 0;
						com_port_timeouts.ReadTotalTimeoutConstant = 0;
						com_port_timeouts.WriteTotalTimeoutMultiplier = 0;
						com_port_timeouts.WriteTotalTimeoutConstant = 0;
						SetCommTimeouts(ekey[devnum].hCom, &com_port_timeouts);
						// reset key and retrieve its serial number
						// send reset command - first time just to stop key sending other packets
						TransmitCommChar(ekey[devnum].hCom, 3);	// ctrl-C
						Sleep(10); 						// wait (10 ms) for key to stop sending other packets
						PurgeComm(ekey[devnum].hCom, PURGE_RXCLEAR); // flush input buffer
						// resend reset command - to retrieve serial number frame
						TransmitCommChar(ekey[devnum].hCom, 3);	// ctrl-C
						// retrieve serial number frame
						read_attempt = 0;
						do
						{
							// read next packet
							ReadFile(ekey[devnum].hCom, input_bytes, PACKET_LEN, &input_count, &ekey[devnum].com_overlapped);
							GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &input_count, TRUE);
							LogFrame(ekey[devnum].com_port, input_bytes, input_count);
						}
						while((++read_attempt < READ_ATTEMPTS) && (strncmp(input_bytes, "* S!", 4) != 0));
						if(read_attempt < READ_ATTEMPTS)
						{
							// check serial number frame against list of specified devices
							listnum = 0;
							found = FALSE;
							while(!found && (listnum < ekey_list_count))
							{
								if(strncmp(input_bytes, ekey_list[listnum].serial_frame, PACKET_LEN) == 0)
								{
									// match found - record device details (already have COM port and read/write 'file' parameters).
									found = TRUE;
									ekey[devnum].ekey_list_num = listnum;
									// convert serial number to binary
									pem64_decode_bytes(ekey_list[listnum].serial_number, 16, ekey[devnum].serial_num);
									// convert master key to binary
									pem64_decode_bytes(ekey_list[listnum].master_key, 44, ekey[devnum].master_key);
									// set session key to zero
									for(i=0;i<32;i++) ekey[devnum].session_key[i] = 0;
									// re-key long-term key
									if(LongTermKey(devnum))
									{
										// re-key session key
										if(SessionKey(devnum))
										{
											// success
											devnum++;
										}
										else LogError(ekey[devnum].com_port, "Failed to re-key session key");
									}
									else LogError(ekey[devnum].com_port, "Failed to re-key long-term key");
								}
								listnum++;
							}
							if(!found) LogError(ekey[devnum].com_port, "Unable to match Serial Number Frame to listed devices");
						} // if(read_attempt < READ_ATTEMPTS)
						else LogError(ekey[devnum].com_port, "Failed to retrieve Serial Number Frame");
					} // if (hCom != INVALID_HANDLE_VALUE)
				} // if(strstr(.) != NULL)
			} // if(SetupDiGetDeviceInterfaceDetail(.))
			HeapFree(GetProcessHeap(), 0, DevIntfDetailData);
			// Continue looping
			SetupDiEnumDeviceInterfaces(hDevInfo, NULL, &GUID_DEVINTERFACE_USB_DEVICE, ++dwMemberIdx, &DevIntfData);
		} // while(GetLastError() != ERROR_NO_MORE_ITEMS)
		SetupDiDestroyDeviceInfoList(hDevInfo);
	} // if(hDevInfo != INVALID_HANDLE_VALUE)

	return devnum;
}

void WritePool(BYTE* entropy, int e_count)
{
	int		i;

	// set semaphore flag for write
	pool_write_flag = TRUE;
	// wait until semaphore flag for read is clear
	while(pool_read_flag);
	// add new entropy to pool
	for(i=0;i<e_count;i++)
	{
		pool[pool_write_index++] = entropy[i];
		if(pool_write_index >= POOLSIZE) pool_write_index = 0;
	}
	// update count of available bytes
	pool_bytes_available += e_count;
	if(pool_bytes_available > POOLSIZE) pool_bytes_available = POOLSIZE;
	// clear semaphore flag for write
	pool_write_flag = FALSE;
}

void EntropyFrame(int devnum)
{
	EKeySkein   skein;
	int		sub_code;
	BYTE	rx_data[40];
	int		rx_count;
	BYTE	encryption_key[32];
	BYTE	entropy[32];
	char	errmsg[100];
	int		i;

	// check sub-code matches expected sequence number
	sub_code = pem64_decode_12bits(ekey[devnum].input_bytes + 4);
	if(sub_code != ekey[devnum].sequence_num)
	{
		sprintf(errmsg, "Missing Entropy frames with sequence numbers %d..%d", ekey[devnum].sequence_num, sub_code - 1);
		LogError(ekey[devnum].com_port, errmsg);
    }
	// update expected sequence number for next frame
	ekey[devnum].sequence_num = sub_code + 1;
	// decode to get binary data load
	rx_count = pem64_decode_bytes(ekey[devnum].input_bytes + 6, 48, rx_data);
	if(rx_count == 32)
	{
		// calculate data encryption key
		PrepareSkein(&skein, ekey[devnum].serial_num, ekey[devnum].session_key, EKEY_SKEIN_PERSONALISATION_EES);
		Skein_256_Update(&skein, (BYTE*) ekey[devnum].input_bytes + 4, 2);
		Skein_256_Final(&skein, encryption_key);
		// Exclusive-OR received data with the encryption key to recover the transmitted entropy
		for(i=0;i<32;i++) entropy[i] = rx_data[i] ^ encryption_key[i];
		// Write entropy to output data file
		LogData(entropy, 32);
		// Add new entropy to pool
		WritePool(entropy, 32);
	} // if(rx_count == 32)
	else LogError(ekey[devnum].com_port, "Incorrect number of bytes of entropy in frame");
}

DWORD WINAPI EkeyPool(void* x)
{
	int		devnum;

	// initialise pool pointers and semaphore flags
	pool_write_index = 0;
	pool_read_index = 0;
	pool_bytes_available = 0;
	pool_write_flag = FALSE;
	pool_read_flag = FALSE;

	// initiate reads from all the successfully connected entropy keys
	for(devnum = 0; devnum < keys_found; devnum++)
		ReadFile(ekey[devnum].hCom, ekey[devnum].input_bytes, PACKET_LEN, &ekey[devnum].input_count, &ekey[devnum].com_overlapped);

	while(!stop_thread)
	{
		// wait until one of the entropy keys sends a frame
		devnum = WaitForMultipleObjects(keys_found, com_event, FALSE, INFINITE) - WAIT_OBJECT_0;
		// read the frame
		GetOverlappedResult(ekey[devnum].hCom, &ekey[devnum].com_overlapped, &ekey[devnum].input_count, FALSE);
		// check frame structure and MAC
		if(ValidMAC(devnum, ekey[devnum].input_bytes, ekey[devnum].input_count))
		{
			// process any Entropy or Out-of-Sequence-Numbers (Re-Key Me) frames
			// other frames just get written to logfile
			switch(ekey[devnum].input_bytes[2])
			{
				case 'E' : // Entropy frame
						   EntropyFrame(devnum);
						   // check expected sequence number for next Entropy frame to see if need new session key
						   if(ekey[devnum].sequence_num >= 4096) SessionKey(devnum);
						   break;
				case 'k' : // Out-of-Sequence-Numbers (Re-Key Me) frame
						   LogFrame(ekey[devnum].com_port, ekey[devnum].input_bytes, ekey[devnum].input_count);
						   SessionKey(devnum);
						   break;
				default :  // All other frame types
						   LogFrame(ekey[devnum].com_port, ekey[devnum].input_bytes, ekey[devnum].input_count);
			}
		} // if(ValidMAC(devnum, ekey[devnum].input_bytes, ekey[devnum].input_count))
		else LogError(ekey[devnum].com_port, "Invalid frame structure or MAC");
		// initiate a new read from that key
		ReadFile(ekey[devnum].hCom, ekey[devnum].input_bytes, PACKET_LEN, &ekey[devnum].input_count, &ekey[devnum].com_overlapped);
	}

	return 0;
}

//---------------------------------------------------------------------------
// ekey_connect() locates any entropy keys present on the system (plugged in),
// matches them against the list of keys given in ekey_list.txt,
// and attempts to reset these keys and re-key their long-term and session keys.
// A separate thread is then launched which continually retrieves the entropy
// generated by the keys and writes it into an entropy pool.
// The return value is the number of entropy keys that have been successfully
// located and initialised.
//
int ekey_connect()
{
	// load entropy key details and pre-calculate serial number frames
	LoadKeyDetails();
	// seed random number generator
	srand(time(NULL));
	// locate any entropy keys currently connected, and re-key the long-term and session keys
	keys_found = FindKeys();
	// launch thread to read in entropy, update pool, and re-key the session keys as necessary
	if(keys_found > 0)
	{
		stop_thread = FALSE;
		ekey_thread = CreateThread(NULL, 0, EkeyPool, NULL, 0, NULL);
	}
	return keys_found;
}

//---------------------------------------------------------------------------
// ekey_disconnect() terminates the execution of the thread launched by
// ekey_connect()
//
void ekey_disconnect()
{
	int		devnum;

	// halt thread reading in data from entropy keys
	stop_thread = TRUE;
	// wait a bit then close thread handle
	Sleep(10);
	if(ekey_thread != NULL) CloseHandle(ekey_thread);
	// close handles to 'files' and Events used for communicating with the keys
	for(devnum = 0; devnum < MAX_KEYS; devnum++)
	{
		if(com_event[devnum] != NULL) CloseHandle(com_event[devnum]);
		if(ekey[devnum].hCom != NULL) CloseHandle(ekey[devnum].hCom);
    }
}

//---------------------------------------------------------------------------
// ekey_get_bytes_available() returns the number of bytes of entropy currently
// available in the pool.
//
int ekey_get_bytes_available()
{
	return pool_bytes_available;
}

//---------------------------------------------------------------------------
// ekey_get_bytes() gets a specified number of random bytes.
// 	 buffer		   : the array to which the retrieved bytes are to be written
//	 reqnum		   : the required number of bytes
//	 ekey_only     : if TRUE only random bytes from the entropy pool are returned
//				   : if FALSE bytes from the system's internal random number
//				     generator will be returned if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the returned bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is the number of random bytes actually written to buffer.
// The bytes will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
// If reqnum <= 0, or if (ekey_only == TRUE) && (reqnum > POOLSIZE) the return
// value is -1.
//
int ekey_get_bytes(BYTE* buffer, int reqnum, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	int		return_value = -1;		// set for fail by default
	int		i;

	*used_ekey_ptr = FALSE;			// set FALSE by default
	// check for valid request
	if((reqnum > 0) && (!ekey_only || (reqnum <= POOLSIZE)))
	{
		return_value = 0;
		// get control of pool
		while(!pool_read_flag)
		{
			// wait until semaphore flag for write is clear
			while(pool_write_flag);
			// set semaphore flag for read
			pool_read_flag = TRUE;
			// recheck semaphore flag for write
			if(pool_write_flag) pool_read_flag = FALSE;
		}
		// read bytes from pool if possible
		if(pool_bytes_available >= reqnum)
		{
			// calculate starting point for read
			pool_read_index = pool_write_index - pool_bytes_available;
			if(pool_read_index < 0) pool_read_index += POOLSIZE;
			// read required number of bytes
			for(i=0;i<reqnum;i++)
			{
				buffer[i] = pool[pool_read_index++];
				if(pool_read_index >= POOLSIZE) pool_read_index = 0;
			}
			// update count of available bytes
			pool_bytes_available -= reqnum;
			*used_ekey_ptr = TRUE;
			return_value = reqnum;
		}
		else // otherwise use system rand() function if permitted
			if(!ekey_only)
			{
				for(i=0;i<reqnum;i++) buffer[i] = rand() % 256;
				return_value = reqnum;
			}
		// clear semaphore flag for read
		pool_read_flag = FALSE;
	} // if((reqnum > 0) && (!ekey_only || (reqnum <= POOLSIZE)))

	return return_value;
}

//---------------------------------------------------------------------------
// ekey_get_bits() gets a specified number of random bits. One bit is written to
// each byte of the output buffer, with each byte set to 0 or 1.
// 	 buffer		   : the array to which the output data are to be written.
//	 reqnum		   : the required number of bit values
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is the number of random bit values written to buffer.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
// If reqnum <= 0, or if reqnum > (8 * POOLSIZE) the return value is -1.
//
int ekey_get_bits(BYTE* buffer, int reqnum, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	int		return_value = -1;		// set for fail by default
	BYTE	byte_buffer[POOLSIZE];
	BYTE	bit_buffer[8];
	int		bytes_required;
	int		bytes_read;
	int		bytenum, bitnum, i;

	*used_ekey_ptr = FALSE;			// set FALSE by default
	// check for valid request
	if((reqnum > 0) && (reqnum <= (8 * POOLSIZE)))
	{
		// get necessary random bytes
		bytes_required = (reqnum + 7) / 8;
		bytes_read = ekey_get_bytes(byte_buffer, bytes_required, ekey_only, used_ekey_ptr);
		if(bytes_read > 0)
		{
			bytenum = bitnum = 0;
			while((bytenum < bytes_read) && (bitnum < reqnum))
			{
				bit_buffer[0] = byte_buffer[bytenum] & 1;
				bit_buffer[1] = byte_buffer[bytenum] & 2;
				bit_buffer[2] = byte_buffer[bytenum] & 4;
				bit_buffer[3] = byte_buffer[bytenum] & 8;
				bit_buffer[4] = byte_buffer[bytenum] & 16;
				bit_buffer[5] = byte_buffer[bytenum] & 32;
				bit_buffer[6] = byte_buffer[bytenum] & 64;
				bit_buffer[7] = byte_buffer[bytenum] & 128;
				bytenum++;
				i = 0;
				while((i < 8) && (bitnum < reqnum))
					buffer[bitnum++] = bit_buffer[i++] ? 1 : 0;
			}
			return_value = bitnum;
		} // if(bytes_read > 0)
		else return_value = bytes_read;
	} // if((reqnum > 0) && (reqnum <= (8 * POOLSIZE)))

	return return_value;
}

//---------------------------------------------------------------------------
// ekey_get_int8/16/32/64 get a single random integer of size 1/2/4/8 bytes
//   op_ptr		   : a pointer to where the output value should be written
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
BOOL ekey_get_int8(__int8* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL	return_value = FALSE;	// set for fail by default

	if(ekey_get_bytes((BYTE*) op_ptr, 1, ekey_only, used_ekey_ptr) == 1)
		return_value = TRUE;

	return return_value;
}
//
BOOL ekey_get_int16(__int16* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	BYTE		byte_buffer[2];
	__int16*	cast_ptr;

	if(ekey_get_bytes(byte_buffer, 2, ekey_only, used_ekey_ptr) == 2)
	{
		cast_ptr = (__int16*) byte_buffer;
		*op_ptr = *cast_ptr;
		return_value = TRUE;
	}

	return return_value;
}
//
BOOL ekey_get_int32(__int32* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	BYTE		byte_buffer[4];
	__int32*	cast_ptr;

	if(ekey_get_bytes(byte_buffer, 4, ekey_only, used_ekey_ptr) == 4)
	{
		cast_ptr = (__int32*) byte_buffer;
		*op_ptr = *cast_ptr;
		return_value = TRUE;
	}

	return return_value;
}
//
BOOL ekey_get_int64(__int64* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	BYTE		byte_buffer[8];
	__int64*	cast_ptr;

	if(ekey_get_bytes(byte_buffer, 8, ekey_only, used_ekey_ptr) == 8)
	{
		cast_ptr = (__int64*) byte_buffer;
		*op_ptr = *cast_ptr;
		return_value = TRUE;
	}

	return return_value;
}

//---------------------------------------------------------------------------
// ekey_get_int_range gets a single random integer with value in the range
// lo_bound..hi_bound, with both bounds included in the range.
//   op_ptr		   : a pointer to where the output value should be written
//	 lo_bound	   : the lower end of the range for the output value
//	 hi_bound	   : the upper end of the range for the output value. The
//					 function actually generates a random offset in the range
//					 0..(hi_bound-lo_bound) and then adds this to lo_bound
//					 to obtain the required output value.
//	 use_mod	   : if TRUE the function generates a 64-bit unsigned random
//					 integer and uses the modulo operator to reduce this to
//					 the required offset range. This approach is similar to that
//					 used by the random() macro (stdlib.h) and is certain to
//					 return a value, but the numbers generated will not have a
//					 flat probability density function.
//				   : if FALSE the function generates a random offset in the
//					 range 0..2^(n-1) where n is the smallest integer such that
//                   the required range will be covered. This approach will
//					 ensure a flat probability density function, but some of the
//					 numbers generated will be too large and it may therefore
//					 require several calls to the function before a number in
//					 the required range is successfully  generated. The
//					 probability of success on each attempt will always be more
//					 than 50%.
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
BOOL ekey_get_int_range(int* op_ptr, int lo_bound, int hi_bound,
							BOOL use_mod, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	__int64		range;
	__int64     r64;
	__int64		offset;
	BYTE		byte_buffer[4];

	range = ((__int64) hi_bound) - ((__int64) lo_bound);
	if(range > 0)  // want random numbers in range 0..range (inclusive)
	{
		if(use_mod)
		{
			if(ekey_get_int64(&r64, ekey_only, used_ekey_ptr))
			{
				*op_ptr = (int)(((__int64) lo_bound) + ((__int64)(((unsigned __int64) r64) % ((unsigned __int64) range + 1))));
				return_value = TRUE;
			}
		} // if(use_mod)
		else // avoid using modulo function
		{
			offset = -1;
			// get necessary random bytes (range == the maximum value required)
			if(range > 16777215) // 2^24 - 1
			{
				// need full 4 bytes
				if(ekey_get_bytes(byte_buffer, 4, ekey_only, used_ekey_ptr) == 4)
				{
					if(range < 2147483648) byte_buffer[3] &= 0x7F;  // 2^31
					if(range < 1073741824) byte_buffer[3] &= 0x3F;	// 2^30
					if(range <  536870912) byte_buffer[3] &= 0x1F;	// 2^29
					if(range <  268435456) byte_buffer[3] &= 0x0F;	// 2^28
					if(range <  134217728) byte_buffer[3] &= 0x07;	// 2^27
					if(range <   67108864) byte_buffer[3] &= 0x03;	// 2^26
					if(range <   33554432) byte_buffer[3] &= 0x01;	// 2^25
//					if(range <   16777216) byte_buffer[3] &= 0x00;	// 2^24
					offset = (((((byte_buffer[3]*256)+byte_buffer[2])*256)+byte_buffer[1])*256)+byte_buffer[0];
				}
			}
			else if(range > 65535) // 2^16 - 1
			{
				// need 3 bytes
				if(ekey_get_bytes(byte_buffer, 3, ekey_only, used_ekey_ptr) == 3)
				{
					if(range <    8388608) byte_buffer[2] &= 0x7F;	// 2^23
					if(range <    4194304) byte_buffer[2] &= 0x3F;	// 2^22
					if(range <    2097152) byte_buffer[2] &= 0x1F;	// 2^21
					if(range <    1048576) byte_buffer[2] &= 0x0F;	// 2^20
					if(range <     524288) byte_buffer[2] &= 0x07;	// 2^19
					if(range <     262144) byte_buffer[2] &= 0x03;	// 2^18
					if(range <     131072) byte_buffer[2] &= 0x01;	// 2^17
//					if(range <      65536) byte_buffer[2] &= 0x00;	// 2^16
					offset = (((byte_buffer[2]*256)+byte_buffer[1])*256)+byte_buffer[0];
				}
			}
			else if(range > 255)  // 2^8 - 1
			{
				// need 2 bytes
				if(ekey_get_bytes(byte_buffer, 2, ekey_only, used_ekey_ptr) == 2)
				{
					if(range <      32768) byte_buffer[1] &= 0x7F;	// 2^15
					if(range <      16384) byte_buffer[1] &= 0x3F;	// 2^14
					if(range <       8192) byte_buffer[1] &= 0x1F;	// 2^13
					if(range <       4096) byte_buffer[1] &= 0x0F;	// 2^12
					if(range <       2048) byte_buffer[1] &= 0x07;	// 2^11
					if(range <       1024) byte_buffer[1] &= 0x03;	// 2^10
					if(range <        512) byte_buffer[1] &= 0x01;	// 2^9
//					if(range <        256) byte_buffer[1] &= 0x00;	// 2^8
					offset = (byte_buffer[1]*256)+byte_buffer[0];
				}
			}
			else
			{
				// need only 1 byte
				if(ekey_get_bytes(byte_buffer, 1, ekey_only, used_ekey_ptr) == 1)
				{
					if(range <        128) byte_buffer[0] &= 0x7F;	// 2^7
					if(range <         64) byte_buffer[0] &= 0x3F;	// 2^6
					if(range <         32) byte_buffer[0] &= 0x1F;	// 2^5
					if(range <         16) byte_buffer[0] &= 0x0F;	// 2^4
					if(range <          8) byte_buffer[0] &= 0x07;	// 2^3
					if(range <          4) byte_buffer[0] &= 0x03;	// 2^2
					if(range <          2) byte_buffer[0] &= 0x01;	// 2^1
//					if(range <          1) byte_buffer[0] &= 0x00;	// 2^0
					offset = byte_buffer[0];
				}
			}
			if(offset > -1)
			{
				*op_ptr = (int)(((__int64) lo_bound) + offset);
				return_value = ((*op_ptr >= lo_bound) && (*op_ptr <= hi_bound));
			}
		} // else avoid using modulo function
	} //if(range > 0)

	return return_value;
}

//---------------------------------------------------------------------------
// ekey_get_float gets a random float value.
// NOTE: This function is specific to the following float representation:
// 		 float has a 4-byte 32-bit format.
// 		 1 (top) bit sign; 8-bit exponent; 23-bit mantissa.
// 		 exponents of all 0's reserved for zeros and subnormal numbers;
// 		 all 1's reserved for infinities and NAN's.
// 		 normal exponent range is 1..254 minus a bias of 127, giving -126..127
// 		 see: http://en.wikipedia.org/wiki/Floating_point
// NOTE: The values returned do NOT have a flat probability density function
//   op_ptr		   : a pointer to where the output value should be written
//   exponent_size : used to limit the number range by restricting the bits
//					 of the exponent. Valid values are 1..8
//					 8 - gives the full number range
//					 7 - limits range of exponent to -63..64 (~10^+-19)
//					 6 - limits range of exponent to -31..32 (~10^+-10)
//					 5 - limits range of exponent to -15..16 (~10^+-5)
//					 4 - limits range of exponent to -7..8   (~10^+-2)
//					 3 - limits range of exponent to -3..4   (~10^+-1)
//					 2 - limits range of exponent to -1..2
//					 1 - limits range of exponent to  0..1
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
BOOL ekey_get_float(float* op_ptr, int exponent_size, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	BOOL		generator_ok;
	WORD		wordvar;

	union r_buffer
	{
		BYTE	byte_buf[4];
		WORD	word_buf[2];
		float	float_buf;
	}           buffer;

	// check valid exponent size
	if((exponent_size > 0) && (exponent_size <= 8))
		do
		{
			generator_ok = (ekey_get_bytes(buffer.byte_buf, 4, ekey_only, used_ekey_ptr) == 4);
			if(generator_ok)
			{
				if(exponent_size == 8)
				{
					// check for valid exponent
					wordvar = buffer.word_buf[1] & 0x7F80;
					if((wordvar != 0) && (wordvar != 0x7F80))
					{
						// exponent ok
						*op_ptr = buffer.float_buf;
						return_value = TRUE;
					}
				}
				else // exponent_size < 8
				{
					switch(exponent_size)
					{
						case 7 : // limit range of exponent to 0..127 + 64 - 127 = -63..64 (~10^+-19)
								 wordvar = (buffer.word_buf[1] & 0xBFFF) + 0x2000;
								 break;
						case 6 : // limit range of exponent to 0..63 + 96 - 127  = -31..32 (~10^+-10)
								 wordvar = (buffer.word_buf[1] & 0x9FFF) + 0x3000;
								 break;
						case 5 : // limit range of exponent to 0..31 + 112 - 127 = -15..16 (~10^+-5)
								 wordvar = (buffer.word_buf[1] & 0x8FFF) + 0x3800;
								 break;
						case 4 : // limit range of exponent to 0..15 + 120 - 127 = -7..8   (~10^+-2)
								 wordvar = (buffer.word_buf[1] & 0x87FF) + 0x3C00;
								 break;
						case 3 : // limit range of exponent to 0..7 + 124 - 127  = -3..4   (~10^+-1)
								 wordvar = (buffer.word_buf[1] & 0x83FF) + 0x3E00;
								 break;
						case 2 : // limit range of exponent to 0..3 + 126 - 127  = -1..2
								 wordvar = (buffer.word_buf[1] & 0x81FF) + 0x3F00;
								 break;
						case 1 : // limit range of exponent to 0..1 + 127 - 127  = 0..1
								 wordvar = (buffer.word_buf[1] & 0x80FF) + 0x3F80;
								 break;
					}
					buffer.word_buf[1] = wordvar;
					*op_ptr = buffer.float_buf;
					return_value = TRUE;
				}
			}
		}
		while(generator_ok && !return_value);

	return return_value;
}

//---------------------------------------------------------------------------

// ekey_get_double gets a random double value.
// NOTE: This function is specific to the following float representation:
// 		 double has an 8-byte 64-bit format.
// 		 1 (top) bit sign; 11-bit exponent; 52-bit mantissa.
// 		 exponents of all 0's reserved for zeros and subnormal numbers;
// 		 all 1's reserved for infinities and NAN's.
// 		 normal exponent range is 1..2046 minus bias of 1023, giving -1022..1023
// 		 see: http://en.wikipedia.org/wiki/Floating_point
// NOTE: The values returned do NOT have a flat probability density function
//   op_ptr		   : a pointer to where the output value should be written
//   exponent_size : used to limit the number range by restricting the bits
//					 of the exponent. Valid values are 1..11
//					 11 - gives the full number range
//                   10 - limits range of exponent to -511..512 (~10^+-154)
//					 9  - limits range of exponent to -255..256 (~10^+-77)
//					 8  - limits range of exponent to -127..128 (~10^+-38)
//					 7  - limits range of exponent to -63..64   (~10^+-19)
//					 6  - limits range of exponent to -31..32   (~10^+-10)
//					 5  - limits range of exponent to -15..16   (~10^+-5)
//					 4  - limits range of exponent to -7..8     (~10^+-2)
//					 3  - limits range of exponent to -3..4     (~10^+-1)
//					 2  - limits range of exponent to -1..2
//					 1  - limits range of exponent to  0..1
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
BOOL ekey_get_double(double* op_ptr, int exponent_size, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default
	BOOL		generator_ok;
	WORD		wordvar;

	union r_buffer
	{
		BYTE	byte_buf[8];
		WORD	word_buf[4];
		double	double_buf;
	}           buffer;

	// check valid exponent size
	if((exponent_size > 0) && (exponent_size <= 11))
		do
		{
			generator_ok = (ekey_get_bytes(buffer.byte_buf, 8, ekey_only, used_ekey_ptr) == 8);
			if(generator_ok)
			{
				if(exponent_size == 11)
				{
					// check for valid exponent
					wordvar = buffer.word_buf[3] & 0x7FF0;
					if((wordvar != 0) && (wordvar != 0x7FF0))
					{
						// exponent ok
						*op_ptr = buffer.double_buf;
						return_value = TRUE;
					}
				}
				else // exponent_size < 11
				{
					switch(exponent_size)
					{
						case 10: // limit range of exponent to 0..1023 + 512 - 1023 = -511..512 (~10^+-154)
								 wordvar = (buffer.word_buf[3] & 0xBFFF) + 0x2000;
								 break;
						case 9 : // limit range of exponent to 0..511 + 768 - 1023  = -255..256 (~10^+-77)
								 wordvar = (buffer.word_buf[3] & 0x9FFF) + 0x3000;
								 break;
						case 8 : // limit range of exponent to 0..255 + 896 - 1023 = -127..128 (~10^+-38)
								 wordvar = (buffer.word_buf[3] & 0x8FFF) + 0x3800;
								 break;
						case 7 : // limit range of exponent to 0..127 + 960 - 1023 = -63..64 (~10^+-19)
								 wordvar = (buffer.word_buf[3] & 0x87FF) + 0x3C00;
								 break;
						case 6 : // limit range of exponent to 0..63 + 992 - 1023  = -31..32 (~10^+-10)
								 wordvar = (buffer.word_buf[3] & 0x83FF) + 0x3E00;
								 break;
						case 5 : // limit range of exponent to 0..31 + 1008 - 1023 = -15..16 (~10^+-5)
								 wordvar = (buffer.word_buf[3] & 0x81FF) + 0x3F00;
								 break;
						case 4 : // limit range of exponent to 0..15 + 1016 - 1023 = -7..8   (~10^+-2)
								 wordvar = (buffer.word_buf[3] & 0x80FF) + 0x3F80;
								 break;
						case 3 : // limit range of exponent to 0..7 + 1020 - 1023  = -3..4   (~10^+-1)
								 wordvar = (buffer.word_buf[3] & 0x807F) + 0x3FC0;
								 break;
						case 2 : // limit range of exponent to 0..3 + 1022 - 1023  = -1..2
								 wordvar = (buffer.word_buf[3] & 0x803F) + 0x3FE0;
								 break;
						case 1 : // limit range of exponent to 0..1 + 1023 - 1023  = 0..1
								 wordvar = (buffer.word_buf[3] & 0x801F) + 0x3FF0;
								 break;
					}
					buffer.word_buf[3] = wordvar;
					*op_ptr = buffer.double_buf;;
					return_value = TRUE;
				}
			}
		}
		while(generator_ok && !return_value);

	return return_value;
}

//---------------------------------------------------------------------------
// ekey_get_double01 gets a random double value in the range 0..1 inclusive.
//   op_ptr		   : a pointer to where the output value should be written
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
BOOL ekey_get_double01(double* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr)
{
	BOOL		return_value = FALSE;	// set for fail by default

	__int64		two28 = 268435456;
	__int64		two56minus1;

	union r_buffer
	{
		BYTE	byte_buf[8];
		__int64	int64_buf;
	}           buffer;

	// double has 52-bit mantissa - so nothing to be gained going beyond this number of bits
	if(ekey_get_bytes(buffer.byte_buf, 7, ekey_only, used_ekey_ptr) == 7)
	{
		buffer.byte_buf[7] = 0; // set top byte zero
		two56minus1 = (two28 * two28) - 1;
		*op_ptr = ((double) buffer.int64_buf) / ((double) two56minus1);
		return_value = TRUE;
	}

	return return_value;
}
//---------------------------------------------------------------------------

