//---------------------------------------------------------------------------
// entropykey.h
//
// Windows DLL for Simtec Entropy Key - www.entropykey.co.uk
//
// Version 1, 21 Oct 2012, Author: Nicolas Lynch-Aird
// Version 2, 22 Oct 2012, Author: Nicolas Lynch-Aird
//			  Ported to MS Visual Studio 2010 Express
// Version 3, 30 Oct 2012, Authors: Nicolas Lynch-Aird, Christopher Mioni
//			  Improvements to solution/project structure and README file
// Version 4, 01 Nov 2012, Author: Nicolas Lynch-Aird
//			  Added use of text file to hold list of entropy key details
//
// This software is released for free use, modification, and distribution. It
// has not been formally tested and is provided with no warranties whatsoever.
//
//---------------------------------------------------------------------------

#if defined(BUILDING_DLL)
#define DllImpExport	__declspec( dllexport )
#elif defined(NOT_USING_DLL)
#define DllImpExport
#else
#define DllImpExport	__declspec( dllimport )
#endif

// MAX_KEYS is the maximum number of entropy keys for which details can be
// defined in the ekey_list[] array.
// POOLSIZE is the size of the entropy pool in bytes
// If either of these numbers are changed then either:
// a) the DLL must then be rebuilt, or
// b) the entropykey.c source code can be included directly into your project using:
//		#define NOT_USING_DLL
//		#include "entropykey.cpp"
//
#define MAX_KEYS 20
#define	POOLSIZE			131072 	// 32 * 4096

#include <Windows.h>
#include <stdio.h>

//---------------------------------------------------------------------------
//
// VARIABLES
// =========
// The following variables should be set by the calling program before
// making any function calls.
//
// ekey_log_fptr: pointer to output file for logging ekey events and all received 
//                frames, except for Entropy frames.
// ekey_dat_fptr: pointer to output file for saving received entropy.
// Default value for both file pointers is NULL.
// Calling program is responsible for opening and closing the associated files.
//
#if defined(BUILDING_DLL)
__declspec( dllexport )	FILE*	ekey_log_fptr = NULL;
__declspec( dllexport )	FILE*	ekey_dat_fptr = NULL;
#elif defined(NOT_USING_DLL)
FILE*	ekey_log_fptr = NULL;
FILE*	ekey_dat_fptr = NULL;
#else
__declspec( dllimport )	FILE*	ekey_log_fptr;
__declspec( dllimport )	FILE*	ekey_dat_fptr;
#endif

//---------------------------------------------------------------------------
//
// FUNCTIONS
// =========

// ekey_connect() locates any entropy keys present on the system (plugged in),
// matches them against the list of keys given in ekey_list.txt,
// and attempts to reset these keys and re-key their long-term and session keys.
// A separate thread is then launched which continually retrieves the entropy
// generated by the keys and writes it into an entropy pool.
// The return value is the number of entropy keys that have been successfully
// located and initialised.
//
DllImpExport int ekey_connect();

// ekey_disconnect() terminates the execution of the thread launched by
// ekey_connect();
//
DllImpExport void ekey_disconnect();

// ekey_get_bytes_available() returns the number of bytes of entropy currently
// available in the pool.
//
DllImpExport int ekey_get_bytes_available();

// ekey_get_bytes() gets a specified number of random bytes.
// 	 buffer		   : the array to which the retrieved bytes are to be written
//	 reqnum		   : the required number of bytes
//	 ekey_only     : if TRUE only random bytes from the entropy pool are returned
//				   : if FALSE bytes from the system's internal random number
//				     generator will be returned if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the returned bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is the number of random bytes actually written to buffer.
// The bytes will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
// If reqnum <= 0, or if (ekey_only == TRUE) && (reqnum > POOLSIZE) the return
// value is -1.
//
DllImpExport int ekey_get_bytes(BYTE* buffer, int reqnum, BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_bits() gets a specified number of random bits. One bit is written to
// each byte of the output buffer, with each byte set to 0 or 1.
// 	 buffer		   : the array to which the output data are to be written.
//	 reqnum		   : the required number of bit values
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is the number of random bit values written to buffer.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
// If reqnum <= 0, or if reqnum > (8 * POOLSIZE) the return value is -1.
//
DllImpExport int ekey_get_bits(BYTE* buffer, int reqnum, BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_int8/16/32/64 get a single random integer of size 1/2/4/8 bytes
//   op_ptr		   : a pointer to where the output value should be written
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
DllImpExport BOOL ekey_get_int8(__int8* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr);
DllImpExport BOOL ekey_get_int16(__int16* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr);
DllImpExport BOOL ekey_get_int32(__int32* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr);
DllImpExport BOOL ekey_get_int64(__int64* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_int_range gets a single random integer with value in the range
// lo_bound..hi_bound, with both bounds included in the range.
//   op_ptr		   : a pointer to where the output value should be written
//	 lo_bound	   : the lower end of the range for the output value
//	 hi_bound	   : the upper end of the range for the output value. The
//					 function actually generates a random offset in the range
//					 0..(hi_bound-lo_bound) and then adds this to lo_bound
//					 to obtain the required output value.
//	 use_mod	   : if TRUE the function generates a 64-bit unsigned random
//					 integer and uses the modulo operator to reduce this to
//					 the required offset range. This approach is similar to that
//					 used by the random() macro (stdlib.h) and is certain to
//					 return a value, but the numbers generated will not have a
//					 flat probability density function.
//				   : if FALSE the function generates a random offset in the
//					 range 0..2^(n-1) where n is the smallest integer such that
//                   the required range will be covered. This approach will
//					 ensure a flat probability density function, but some of the
//					 numbers generated will be too large and it may therefore
//					 require several calls to the function before a number in
//					 the required range is successfully  generated. The
//					 probability of success on each attempt will always be more
//					 than 50%.
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
DllImpExport BOOL ekey_get_int_range(int* op_ptr, int lo_bound, int hi_bound,
						BOOL use_mod, BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_float gets a random float value.
// NOTE: This function is specific to the following float representation:
// 		 float has a 4-byte 32-bit format.
// 		 1 (top) bit sign; 8-bit exponent; 23-bit mantissa.
// 		 exponents of all 0's reserved for zeros and subnormal numbers;
// 		 all 1's reserved for infinities and NAN's.
// 		 normal exponent range is 1..254 minus a bias of 127, giving -126..127
// 		 see: http://en.wikipedia.org/wiki/Floating_point
// NOTE: The values returned do NOT have a flat probability density function
//   op_ptr		   : a pointer to where the output value should be written
//   exponent_size : used to limit the number range by restricting the bits
//					 of the exponent. Valid values are 1..8
//					 8 - gives the full number range
//					 7 - limits range of exponent to -63..64 (~10^+-19)
//					 6 - limits range of exponent to -31..32 (~10^+-10)
//					 5 - limits range of exponent to -15..16 (~10^+-5)
//					 4 - limits range of exponent to -7..8   (~10^+-2)
//					 3 - limits range of exponent to -3..4   (~10^+-1)
//					 2 - limits range of exponent to -1..2
//					 1 - limits range of exponent to  0..1
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
DllImpExport BOOL ekey_get_float(float* op_ptr, int exponent_size,
										BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_double gets a random double value.
// NOTE: This function is specific to the following float representation:
// 		 double has an 8-byte 64-bit format.
// 		 1 (top) bit sign; 11-bit exponent; 52-bit mantissa.
// 		 exponents of all 0's reserved for zeros and subnormal numbers;
// 		 all 1's reserved for infinities and NAN's.
// 		 normal exponent range is 1..2046 minus bias of 1023, giving -1022..1023
// 		 see: http://en.wikipedia.org/wiki/Floating_point
// NOTE: The values returned do NOT have a flat probability density function
//   op_ptr		   : a pointer to where the output value should be written
//   exponent_size : used to limit the number range by restricting the bits
//					 of the exponent. Valid values are 1..11
//					 11 - gives the full number range
//                   10 - limits range of exponent to -511..512 (~10^+-154)
//					 9  - limits range of exponent to -255..256 (~10^+-77)
//					 8  - limits range of exponent to -127..128 (~10^+-38)
//					 7  - limits range of exponent to -63..64   (~10^+-19)
//					 6  - limits range of exponent to -31..32   (~10^+-10)
//					 5  - limits range of exponent to -15..16   (~10^+-5)
//					 4  - limits range of exponent to -7..8     (~10^+-2)
//					 3  - limits range of exponent to -3..4     (~10^+-1)
//					 2  - limits range of exponent to -1..2
//					 1  - limits range of exponent to  0..1
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
DllImpExport BOOL ekey_get_double(double* op_ptr, int exponent_size,
										BOOL ekey_only, BOOL* used_ekey_ptr);

// ekey_get_double01 gets a random double value in the range 0..1 inclusive.
//   op_ptr		   : a pointer to where the output value should be written
//	 ekey_only     : if TRUE only random bytes from the entropy pool are used
//				   : if FALSE bytes from the system's internal random number
//				     generator will be used if there are not enough bytes
//				     available in the entropy pool.
//   used_ekey_ptr : pointer to a boolean; set to TRUE if the random bytes
//					 were taken from the entropy pool, FALSE otherwise.
// The return value is TRUE if the function succeeded, FALSE otherwise.
// The bits will either be all from the pool or all from the system's internal
// random number generator (depending on ekey_only) - they will not be a mixture
//
DllImpExport BOOL ekey_get_double01(double* op_ptr, BOOL ekey_only, BOOL* used_ekey_ptr);







